__version__ = '1.11.1'

import os
import re
import sys
import atexit
import signal
import getopt
import collections

from .isddc_dict import IsddcDict as idict
from .isddc_command import IsddcCommand as icmd

from . import isddc_baseclass as ibase
from . import isddc_time as itime
from . import isddc_filesystem as ifs
from . import isddc_log as ilog
from . import isddc_pool as ipool
from . import isddc_decorator as idec
from . import isddc_crypto as icrypto
from . import isddc_obfuscer as iobfuscer

iapi: 'isddc api'
issh: 'isddc ssh'
ikube: 'isddc kube'
ikafka: 'isddc kafka'
imongo: 'isddc mongo'
imysql: 'isddc mysql'
iinflux: 'isddc influx'
idbpool: 'isddc dbpool'

# Isddc Config.
inf: idict

linux: bool = True if os.name == 'posix' else False

if linux and not icmd('fdisk -l').output:
    container = True
else:
    container = False


def __init__(config: dict):
    setattr(sys.modules[__name__], 'inf', config)

    for iname, iconf in config.items():
        if iname[0] != 'i':
            continue

        try:
            imodule = Import.module(f'{__name__}.isddc_{iname[1:]}')
        except ModuleNotFoundError:
            continue

        init: dict = iconf.pop('init', {})

        for name, params in iconf.items():
            for item in init.items():
                params.setdefault(*item)

            iobj = imodule.__init__(params)

            if not hasattr(imodule, '__default__'):
                setattr(imodule, '__default__', iobj)

            setattr(imodule, name, iobj)

        setattr(sys.modules[__name__], iname, imodule)


class DictInterProcess(metaclass=ibase.FuncMode):
    """Dictionary internal processing
    Used to batch process values inside a dictionary.
    :param data: A dict.
    :param func: Your handler functions, ps:
        def func(name, value):
            if name == 'xxx':
                return process(value)
    """

    def __init__(self, data: dict, func=None):
        self.data = data
        self.func = func

    def __call__(self) -> dict:
        return self.one(self.data)

    def one(self, data: dict):
        for name, value in data.items():
            data[name] = self.two(name, value)

        return data

    def two(self, name=None, value={}):
        value = self.func(name, value) or value

        if isinstance(value, dict):
            return self.one(value)

        if isinstance(value, collections.Iterable) \
                and not isinstance(value, str):
            return [self.two(value=v) for v in value]

        return value


class BootParams(metaclass=ibase.FuncMode):
    """Get boot parameters.
    Boot parameters with default values
    are available through **default_params.
    """
    for_later = {}

    def __init__(self, *params, **default_params):
        self.short_opts = ''
        self.long_opts = []

        for x in params:
            if len(x) == 1 or x[-1] == ':':
                self.short_opts += x
            else:
                self.long_opts.append(x)

        for x in default_params:
            if len(x) == 1:
                self.short_opts += f'{x}:'
            else:
                self.long_opts.append(f'{x}=')

        BootParams.for_later = default_params

    def __call__(self) -> dict:
        params: list = getopt.getopt(
            args=sys.argv[1:],
            shortopts=self.short_opts,
            longopts=self.long_opts
        )[0]

        for k, v in params:
            if k[:2] == '--':
                BootParams.for_later[k[2:]] = v
            elif k[0] == '-':
                BootParams.for_later[k[1:]] = v
            else:
                BootParams.for_later = v

        return BootParams.for_later


class Import:

    @classmethod
    def module(cls, module: str):
        __import__(module)
        return sys.modules[module]

    @classmethod
    def name(cls, name: str):
        module_, _, name = name.rpartition('.')
        module_ = cls.module(module_)
        return getattr(module_, name)

    @classmethod
    def object(cls, class_path: str, *a, **kw):
        return cls.name(class_path)(*a, **kw)

    @classmethod
    def try_import(cls, import_str, default=None):
        try:
            return cls.module(import_str)
        except ImportError:
            return default


class OverFunc:
    """Functions in `__over_func__` are
     executed at the end of the program."""
    __over_func__ = []

    def __init__(self, func, *a, **kw):
        """Add a function that will execute
        after the main program is finished"""
        self.__over_func__.append((func, a, kw))

    @staticmethod
    @atexit.register
    def __exec__():
        """Will execute when the program exits normally."""
        for func, a, kw in OverFunc.__over_func__:
            try:
                func(*a, **kw)
            except Exception as e:
                ilog.simple.error(f'OverFunc.{func.__name__} error: {e}')

    # Trigger `__exec__` function after being killed.
    signal.signal(signal.SIGTERM, lambda signum, stack_frame: exit(1))


def hump(name: str) -> str:
    if '_' in name:
        name: str = ''.join(_.capitalize() for _ in name.split('_'))
    if '-' in name:
        name: str = ''.join(_.capitalize() for _ in name.split('-'))
    return name[0].upper() + name[1:]


def underline(name: str) -> str:
    for _ in re.findall(r'([A-Z]+)[A-Z]', name):
        name = name.replace(_, _.capitalize())

    result: list = re.findall(r'[A-Z][a-z]+|[a-z]+|[A-Z]\d', name)
    return '_'.join(_.lower() for _ in result)


def get_caller(level: int = 1) -> str:
    back = sys._getframe().f_back

    for _ in range(level):
        back = back.f_back

    return back.f_code.co_name


def avg(numbers: list) -> float:
    return sum(numbers) / len(numbers)


def prt(*args, color: int = 31, font: int = 0,
        end: str = '\n', sep: str = ' ', file: open = None):
    """
    Custom print font and color.

    Color:
        31-red 32-green 34-blue 37-Gray 38-black
    Font:
        0-normal 1-bold
    sep:
        string inserted between values, default a space.
    """
    info: str = sep.join(str(i) for i in args)
    format: str = f'\033[{font};{color};0m{info}\033[0m'
    print(format, end=end, file=file)
