import re
import abc
import time
import functools
import traceback
import isddc as i

from . import itime
from . import ibase
from . import isddc_log as ilog


class BaseDecorator(metaclass=abc.ABCMeta):

    def __call__(self, func):
        @functools.wraps(func)
        def inner(*a, **kw):
            return self.core(func, *a, **kw)

        return inner

    @abc.abstractmethod
    def core(self, func, *a, **kw):
        ...


class TryExcept(BaseDecorator):

    def __init__(
            self,
            exc_type: type or tuple = Exception,
            *,
            mark: str = None,
            ignore: bool = False,
            exc_return=None,
            exc_callback=None
    ):
        self.mark = mark
        self.ignore = ignore
        self.exc_type = exc_type
        self.exc_return = exc_return
        self.exc_callback = exc_callback

    def core(self, func, *a, **kw):
        try:
            return func(*a, **kw)
        except self.exc_type as e:
            if self.exc_callback:
                return self.exc_callback(*a, **kw)

            if not self.ignore:
                try:
                    exc_info: str = self.analysis_exception(func, e)
                    ilog.simple.error(exc_info)
                except Exception as ee:
                    ilog.simple.error(f'TryExceptError: {ee}')

        return self.exc_return

    def analysis_exception(self, func, err: Exception) -> str:
        exc_name: str = type(err).__name__
        exc_position: str = re.search(
            r'File.+', traceback.format_exc(limit=-1)).group()

        if self.mark is None:
            self.mark: str = f'{func.__globals__["__name__"]}.{func.__qualname__}'

        return f'[{self.mark}] [{exc_position}] {exc_name}: {str(err)}'


class Retry(BaseDecorator):

    def __init__(
            self,
            mark: str = None,
            *,
            count: int = None,
            cycle: int = 0,
            retry_exc: type or tuple = Exception,
    ):
        self.mark = mark
        self.count = count
        self.cycle = cycle
        self.retry_exc = retry_exc

    def core(self, func, *a, **kw):
        count, max_count = 0, self.count or "N"

        while True:
            try:
                return func(*a, **kw)
            except self.retry_exc as e:
                try:
                    count += 1
                    exc_info: str = TryExcept.analysis_exception(self, func, e)
                    ilog.simple.warning(f'[try:{count}/{max_count}] {exc_info}')
                except Exception as ee:
                    ilog.simple.error(f'RetryError: {ee}')

                if self.count and count == self.count:
                    raise e

            time.sleep(self.cycle)


class WhileTrue(BaseDecorator):

    def __init__(
            self,
            cycle: int or float = 0,
            *,
            minus_run_time: bool = False,
            wait_first: bool = False
    ):
        self.cycle = cycle
        self.minus_run_time = minus_run_time
        self.wait_first = wait_first

    def core(self, func, *a, **kw):
        cycle: int = self.cycle

        if not self.wait_first:
            ret, cycle = self.call(func, *a, **kw)

        while True:
            time.sleep(cycle)
            ret, cycle = self.call(func, *a, **kw)

            if ret == '_break_':
                break
            if ret == '_continue_':
                continue

    def call(self, func, *a, **kw):
        if self.minus_run_time and self.cycle:
            start = itime.Now.float
            ret = func(*a, **kw)
            end = itime.Now.float
            cycle = self.cycle - (end - start)
        else:
            ret = func(*a, **kw)
            cycle = self.cycle

        return ret, cycle


class RunTime(BaseDecorator):

    def __init__(self, mark: str = None):
        self.mark = mark

    def core(self, func, *a, **kw):
        start = time.time()
        func(*a, **kw)
        end = time.time()

        mark: str = self.mark or i.hump(func.__qualname__)
        exec_time: str = itime.second2time(end - start)
        ilog.simple.info(f'{mark}: {exec_time}')


class TestFuncSpeed(BaseDecorator):

    def __init__(self, count: int, *, keep: int = 7):
        self.count = count
        self.keep = keep

    def core(self, func, *a, **kw):
        speeds = []

        for _ in range(self.count):
            start = time.time()
            func(*a, **kw)
            end = time.time()
            speeds.append(end - start)

        for action in min, max, i.avg, sum:
            result = round(action(speeds), self.keep)
            print(f'{action.__name__}: {result}')


class BeforeFunc(BaseDecorator):

    def __init__(self, before_func, *, enable: bool = True):
        self.before_func = before_func
        self.enable = enable

    def core(self, func, *a, **kw):
        if not self.enable:
            return func(*a, **kw)

        try:
            result = self.before_func(*a, **kw)
        except TypeError:
            result = self.before_func()

        if not isinstance(result, dict):
            return func(*a, **kw)

        a: list = list(a)

        if isinstance(func, ibase.FuncMode):
            co_varnames: tuple = func.__init__.__code__.co_varnames
        else:
            co_varnames: tuple = func.__code__.co_varnames

        if 'a' in co_varnames and 'kw' in co_varnames or 'self' in co_varnames:
            kw.update(result)
            return func(*a, **kw)

        for idx, param in enumerate(co_varnames):
            if param in result:
                try:
                    a[idx] = result[param]
                except IndexError:
                    kw[param] = result[param]

        return func(*a, **kw)


class AfterFunc(BaseDecorator):

    def __init__(
            self,
            after_func,
            *,
            use_raw_params: bool = False,
            enable: bool = True
    ):
        self.after_func = after_func
        self.use_raw_params = use_raw_params
        self.enable = enable

    def core(self, func, *a, **kw):
        raw_result = func(*a, **kw)

        if not self.enable:
            return raw_result

        if self.use_raw_params:
            result = self.after_func(*a, **kw)
        elif raw_result is None:
            try:
                result = self.after_func(a[0])
            except (IndexError, TypeError):
                result = self.after_func()
        else:
            try:
                result = self.after_func(a[0], *raw_result if isinstance(
                    raw_result, tuple) else raw_result)
            except (IndexError, TypeError):
                result = self.after_func(*raw_result if isinstance(
                    raw_result, tuple) else raw_result)

        return raw_result if result is None else result
