import sys
import time
import threading

from pymysql import cursors
from pymysql.connections import Connection

from . import idec
from . import idict

__default__: Connection
__lock__ = threading.Lock()


@idec.TryExcept(Exception, mark='InitMySQL')
@idec.Retry('InitMySQL', count=10, cycle=10)
def __init__(params: dict):
    return Connection(**params)


def sql(s: str, *, one: bool = False, dtype=dict, conn: Connection = None):
    conn = conn or __default__
    conn.ping(reconnect=True)

    cur = conn.cursor(Transaction.dtypes[dtype])

    with __lock__:
        result: int = cur.execute(s)
        conn.commit()

    if s.lstrip()[:6].upper() in ['INSERT', 'UPDATE', 'DELETE']:
        return result

    queryset = cur.fetchall()

    if dtype is dict:
        queryset = Transaction.upgrade_dtype(queryset)

    cur.close()

    return queryset[0] if one else queryset


class Transaction:
    dtypes = {
        tuple: cursors.Cursor,
        dict: cursors.DictCursor
    }

    def __init__(self, dtype=dict, *, conn: Connection = None):
        self.conn = conn or __default__
        conn.ping(reconnect=True)
        self.cur = self.conn.cursor(self.dtypes[dtype])
        self.dtype = dtype

    def __call__(self, s: str) -> int:
        return self.cur.execute(s)

    def __enter__(self):
        self.__lock__.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type or exc_val or exc_tb:
            self.conn.rollback()
        self.conn.commit()
        self.__lock__.release()
        self.cur.close()

    def __getattribute__(self, name):
        try:
            result = super().__getattribute__(name)
        except AttributeError:
            result = getattr(self.cur, name)

        return result

    def __invert__(self):
        return self.fetchone

    @staticmethod
    def upgrade_dtype(data: list = None) -> list:
        """dict -> Dict"""
        return list(map(lambda x: idict(**x), data))

    @property
    def fetchone(self):
        queryset = self.cur.fetchone()

        if self.dtype is tuple:
            return queryset

        return idict(**queryset)

    @property
    def fetchall(self):
        queryset = self.cur.fetchall()

        if self.dtype is tuple:
            return queryset

        return self.upgrade_dtype(queryset)

    @property
    def last_insert_id(self) -> int:
        """
        Returns the id of the data
        that was last inserted by `self.conn`
        """
        self('SELECT LAST_INSERT_ID()')

        if self.dtype is tuple:
            return self.fetchone[0]

        return self.fetchone['LAST_INSERT_ID()']


def __health_examination__(cycle: int = 60 * 10):
    """
    Open a thread, Periodically check the database link status,
    and if it is disconnected, it will restart the connection.
    """

    def __(conns: 'Not param' = []):
        for name in sys.modules[__name__].__dir__():
            value = getattr(sys.modules[__name__], name)
            if name != '__default__' and isinstance(value, Connection):
                conns.append(value)

        while True:
            time.sleep(cycle)
            for conn in conns:
                conn.ping(reconnect=True)

    threading.Thread(target=__, name='PingMySQL', daemon=True).start()
